@@ -14,7 +14,10 @@
 #define SHOW_URL
 #endif
 
-static struct stat bookmarks_st;
+int bookmarks_codepage=0;
+static int can_write_bookmarks=0;	/* global flag if we can write bookmarks */
+
+unsigned char bookmarks_file[MAX_STR_LEN]="";
 
 static void *bookmark_new_item(void *);
 static unsigned char *bookmark_type_item(struct terminal *, void *, int);
@@ -24,7 +27,6 @@ static void bookmark_copy_item(void *, void *);
 static void bookmark_goto_item(struct session *, void *);
 static void *bookmark_default_value(struct session*, unsigned char);
 static void *bookmark_find_item(void *start, unsigned char *str, int direction);
-static void save_bookmarks(struct session *ses);
 
 struct list bookmarks={&bookmarks,&bookmarks,0,-1,NULL};
 
@@ -66,6 +68,7 @@ static struct list_description bookmark_ld=
 	bookmark_find_item,
 	&bookmark_search_history,
 	0,		/* this is set in init_bookmarks function */
+	60,  /* width of main window */
 	15,  /* # of items in main window */
 	T_BOOKMARK,
 	T_BOOKMARKS_ALREADY_IN_USE,
@@ -73,9 +76,8 @@ static struct list_description bookmark_ld=
 	T_DELETE_BOOKMARK,
 	T_GOTO,
 	bookmark_goto_item,	/* FIXME: should work (URL in UTF8), but who knows? */
-	save_bookmarks,
 
-	NULL,NULL,0,0,  /* internal vars */
+	0,0,0,0,  /* internal vars */
 	0, /* modified */
 	NULL,
 	NULL,
@@ -108,6 +110,7 @@ static void free_bookmarks(void)
 /* called before exiting the links */
 void finalize_bookmarks(void)
 {
+	save_bookmarks();
 	free_bookmarks();
 }
 
@@ -134,7 +137,7 @@ static void *bookmark_default_value(struct session *ses, unsigned char type)
 			struct conv_table* ct;
 			
 			ct=get_translation_table(ses->term->spec->charset,bookmark_ld.codepage);
-			zelena->url=convert_string(ct,txt,strlen(cast_const_char txt),NULL);
+			zelena->url=convert_string(ct,txt,strlen(txt),NULL);
 			clr_white(zelena->url);
 		}
 		else
@@ -145,7 +148,7 @@ static void *bookmark_default_value(struct session *ses, unsigned char type)
 		struct conv_table* ct;
 		
 		ct=get_translation_table(ses->term->spec->charset,bookmark_ld.codepage);
-		zelena->title=convert_string(ct,txt,strlen(cast_const_char txt),NULL);
+		zelena->title=convert_string(ct,txt,strlen(txt),NULL);
 		clr_white(zelena->title);
 	}
 
@@ -258,7 +261,7 @@ static void bookmark_edit_done(void *data)
 		
 		mem_free(item->title);
 		ct=get_translation_table(s->dlg->win->term->spec->charset,bookmark_ld.codepage);
-		item->title=convert_string(ct,title,strlen(cast_const_char title),NULL);
+		item->title=convert_string(ct,title,strlen(title),NULL);
 		clr_white(item->title);
 	}
 
@@ -268,12 +271,12 @@ static void bookmark_edit_done(void *data)
 		
 		mem_free(item->url);
 		ct=get_translation_table(s->dlg->win->term->spec->charset,bookmark_ld.codepage);
-		item->url=convert_string(ct,url,strlen(cast_const_char url),NULL);
+		item->url=convert_string(ct,url,strlen(url),NULL);
 		clr_white(item->url);
 	}
 
 	s->fn(s->dlg,s->data,item,&bookmark_ld);
-	d->udata=NULL;  /* for abort function */
+	d->udata=0;  /* for abort function */
 }
 
 
@@ -318,12 +321,12 @@ static void bookmark_edit_item(struct dialog_data *dlg,void *data,void (*ok_fn)(
 		struct conv_table* ct;
 		
 		ct=get_translation_table(bookmark_ld.codepage,dlg->win->term->spec->charset);
-		txt=convert_string(ct,item->title,strlen(cast_const_char item->title),NULL);
+		txt=convert_string(ct,item->title,strlen(item->title),NULL);
 		clr_white(txt);
 		safe_strncpy(title,txt,MAX_STR_LEN);
 		mem_free(txt);
 
-		txt=convert_string(ct,item->url,strlen(cast_const_char item->url),NULL);
+		txt=convert_string(ct,item->url,strlen(item->url),NULL);
 		clr_white(txt);
 		safe_strncpy(url,txt,MAX_STR_LEN);
 		mem_free(txt);
@@ -433,13 +436,13 @@ static unsigned char *bookmark_type_item(struct terminal *term, void *data, int
 #endif
 	if (!x&&!((item->type)&1))
 	{
-		add_to_strn(&txt,cast_uchar "   (");
+		add_to_strn(&txt,"   (");
 		if (item->url)add_to_strn(&txt,item->url);
-		add_to_strn(&txt,cast_uchar ")");
+		add_to_strn(&txt,")");
 	}
 
 	table=get_translation_table(bookmark_ld.codepage,term->spec->charset);
-	txt1=convert_string(table,txt,strlen(cast_const_char txt),NULL);
+	txt1=convert_string(table,txt,strlen(txt),NULL);
 	clr_white(txt1);
 	mem_free(txt);
 	return txt1;
@@ -475,7 +478,7 @@ static int substr_utf8(unsigned char *string, unsigned char *substr)
 	int r;
 	string = unicode_upcase_string(string);
 	substr = unicode_upcase_string(substr);
-	r = !!strstr(cast_const_char string, cast_const_char substr);
+	r = !!strstr(string, substr);
 	mem_free(string);
 	mem_free(substr);
 	return r;
@@ -548,7 +551,7 @@ static void add_bookmark(unsigned char *title, unsigned char *url, int depth)
 		struct conv_table* ct;
 		
 		ct=get_translation_table(bookmarks_codepage,bookmark_ld.codepage);
-		b->title=convert_string(ct,title,strlen(cast_const_char title),dop);
+		b->title=convert_string(ct,title,strlen(title),dop);
 		clr_white(b->title);
 	}
 	
@@ -558,7 +561,7 @@ static void add_bookmark(unsigned char *title, unsigned char *url, int depth)
 		
 		dop->plain=1;
 		ct=get_translation_table(bookmarks_codepage,bookmark_ld.codepage);
-		b->url=convert_string(ct,url,strlen(cast_const_char url),dop);
+		b->url=convert_string(ct,url,strlen(url),dop);
 		clr_white(b->url);
 		dop->plain=0;
 
@@ -589,19 +592,19 @@ static void add_bookmark(unsigned char *title, unsigned char *url, int depth)
 /* Created pre-cooked bookmarks */
 static void create_initial_bookmarks(void)
 {
-	bookmarks_codepage=utf8_table;
-	add_bookmark(cast_uchar "Links",NULL,0);
-	add_bookmark(cast_uchar "English",NULL,1);
-	add_bookmark(cast_uchar "Calibration Procedure",cast_uchar "http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/calibration.html",2);
-	add_bookmark(cast_uchar "Links Homepage",cast_uchar "http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/",2);
-	add_bookmark(cast_uchar "Links Manual",cast_uchar "http://links.twibright.com/user_en.html",2);
-	add_bookmark(cast_uchar "Cesky",NULL,1);
-	add_bookmark(cast_uchar "Kalibracni procedura",cast_uchar "http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/kalibrace.html",2);
-	add_bookmark(cast_uchar "Links: domaci stranka",cast_uchar "http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/index_cz.html",2);
-	add_bookmark(cast_uchar "Manual k Linksu",cast_uchar "http://links.twibright.com/user.html",2);
+	bookmarks_codepage=get_cp_index("8859-2");
+	add_bookmark("Links",NULL,0);
+	add_bookmark("English",NULL,1);
+	add_bookmark("Calibration Procedure","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/calibration.html",2);
+	add_bookmark("Links Homepage","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/",2);
+	add_bookmark("Links Manual","http://links.twibright.com/user_en.html",2);
+	add_bookmark("Cesky",NULL,1);
+	add_bookmark("Kalibracni procedura","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/kalibrace.html",2);
+	add_bookmark("Links: domaci stranka","http://atrey.karlin.mff.cuni.cz/~clock/twibright/links/index_cz.html",2);
+	add_bookmark("Manual k Linksu","http://links.twibright.com/user.html",2);
 }
 
-static void load_bookmarks(struct session *ses)
+static void load_bookmarks(void)
 {
 	unsigned char *buf;
 	long len;
@@ -615,7 +618,6 @@ static void load_bookmarks(struct session *ses)
 	int depth;
 
 	struct document_options dop;
-	int rs;
 	
 	memset(&dop, 0, sizeof(dop));
 	dop.plain=1;
@@ -630,14 +632,15 @@ static void load_bookmarks(struct session *ses)
 	 */
 
 	buf=read_config_file(bookmarks_file);
+	can_write_bookmarks=1;
 	if (!buf){
+		can_write_bookmarks=(errno==ENOENT); /* if open failed and the file exists, don't write initial bookmarks when finishing */
 		create_initial_bookmarks();
 		bookmark_ld.modified=1;
-		save_bookmarks(ses);
 		return;
 	}
 
-	len=strlen(cast_const_char buf);
+	len=strlen(buf);
 
 	p=buf;
 	end=buf+len;
@@ -659,9 +662,9 @@ static void load_bookmarks(struct session *ses)
 		switch (status)
 		{
 			case 0:  /* <dt> or </dl> */
-			if (namelen==2&&!casecmp(name,cast_uchar "dt",2))
+			if (namelen==2&&!casecmp(name,"dt",2))
 				status=1;
-			else if (namelen==3&&!casecmp(name,cast_uchar "/dl",3))
+			else if (namelen==3&&!casecmp(name,"/dl",3))
 			{
 				depth--;
 				if (depth==-1)goto smitec;
@@ -669,13 +672,13 @@ static void load_bookmarks(struct session *ses)
 			continue;
 
 			case 1:   /* find "a" element */
-			if (namelen==1&&!casecmp(name,cast_uchar "a",1))
+			if (namelen==1&&!casecmp(name,"a",1))
 			{
-				if (!(url=get_attr_val(attr,cast_uchar "href")))continue;
+				if (!(url=get_attr_val(attr,"href")))continue;
 				status=2;
 				title=p;
 			}
-			if (namelen==2&&!casecmp(name,cast_uchar "h3",1))
+			if (namelen==2&&!casecmp(name,"h3",1))
 			{
 				status=3;
 				title=p;
@@ -683,7 +686,7 @@ static void load_bookmarks(struct session *ses)
 			continue;
 
 			case 2:   /* find "/a" element */
-			if (namelen!=2||casecmp(name,cast_uchar "/a",2))continue;   /* ignore all other elements */
+			if (namelen!=2||casecmp(name,"/a",2))continue;   /* ignore all other elements */
 			*s=0;
 			add_bookmark(title,url,depth);
 			mem_free(url);
@@ -691,60 +694,41 @@ static void load_bookmarks(struct session *ses)
 			continue;
 
 			case 3:   /* find "/h3" element */
-			if (namelen!=3||casecmp(name,cast_uchar "/h3",2))continue;   /* ignore all other elements */
+			if (namelen!=3||casecmp(name,"/h3",2))continue;   /* ignore all other elements */
 			*s=0;
 			add_bookmark(title,NULL,depth);
 			status=0;
 			depth++;
 			continue;
 		}
+		
 	}
 	if (status==2)mem_free(url);
 smitec:
 	mem_free(buf);
 	d_opt=&dd_opt;
 	bookmark_ld.modified=0;
-
-	EINTRLOOP(rs, stat(cast_const_char bookmarks_file, &bookmarks_st));
-	if (rs)
-		memset(&bookmarks_st, 0, sizeof bookmarks_st);
 }
 
 void init_bookmarks(void)
 {
-	memset(&bookmarks_st, 0, sizeof bookmarks_st);
 	if (!*bookmarks_file)
-		snprintf(cast_char bookmarks_file,MAX_STR_LEN,"%sbookmarks.html",links_home?links_home:(unsigned char*)"");
+		snprintf(bookmarks_file,MAX_STR_LEN,"%sbookmarks.html",links_home?links_home:(unsigned char*)"");
 
-	bookmark_ld.codepage=utf8_table;
-	load_bookmarks(NULL);
+	bookmark_ld.codepage=get_cp_index("utf-8");
+	load_bookmarks();
 }
 
-void reinit_bookmarks(struct session *ses, unsigned char *new_bookmarks_file, int new_bookmarks_codepage)
+void reinit_bookmarks(void)
 {
-	unsigned char *buf;
-	if (test_list_window_in_use(&bookmark_ld, ses->term))
-		return;
-
-	if (!strcmp(cast_const_char bookmarks_file, cast_const_char new_bookmarks_file)) {
-		goto save_only;
-	}
-
-	buf=read_config_file(new_bookmarks_file);
-	if (buf) {
-		mem_free(buf);
-		free_bookmarks();
-		safe_strncpy(bookmarks_file,new_bookmarks_file,MAX_STR_LEN);
-		bookmarks_codepage=new_bookmarks_codepage;
-		load_bookmarks(ses);
-		reinit_list_window(&bookmark_ld);
-	} else {
-		save_only:
-		safe_strncpy(bookmarks_file,new_bookmarks_file,MAX_STR_LEN);
-		bookmarks_codepage=new_bookmarks_codepage;
-		bookmark_ld.modified=1;
-		save_bookmarks(ses);
-	}
+	free_bookmarks();
+	bookmarks.next=&bookmarks;
+	bookmarks.prev=&bookmarks;
+	bookmarks.type=0;
+	bookmarks.depth=-1;
+	bookmarks.fotr=NULL;
+	load_bookmarks();
+	reinit_list_window(&bookmark_ld);
 }
 
 
@@ -792,7 +776,7 @@ static unsigned char *convert_to_entity_string(unsigned char *str)
 }
 
 /* writes bookmarks to disk */
-static void save_bookmarks(struct session *ses)
+void save_bookmarks(void)
 {
 	struct bookmark_list *b;
 	int depth;
@@ -800,14 +784,12 @@ static void save_bookmarks(struct session *ses)
 	struct conv_table *ct;
 	unsigned char *data;
 	int l;
-	int err;
-	int rs;
 
-	if (!bookmark_ld.modified)return;
+	if (!bookmark_ld.modified||!can_write_bookmarks)return;
 	ct=get_translation_table(bookmark_ld.codepage,bookmarks_codepage);
 	data=init_str();
 	l=0;
-	add_to_str(&data, &l, cast_uchar 
+	add_to_str(&data, &l,
 	"<HTML>\n"
 	"<HEAD>\n"
 	"<!-- This is an automatically generated file.\n"
@@ -821,18 +803,18 @@ static void save_bookmarks(struct session *ses)
 	depth=0;
 	foreach(b,bookmarks)
 	{
-		for (a=b->depth;a<depth;a++)add_to_str(&data, &l,cast_uchar "</DL>\n");
+		for (a=b->depth;a<depth;a++)add_to_str(&data, &l,"</DL>\n");
 		depth=b->depth;
 	
 		if ((b->type)&1)
 		{
 			unsigned char *txt, *txt1;
-			txt=convert_string(ct,b->title,strlen(cast_const_char b->title),NULL);
+			txt=convert_string(ct,b->title,strlen(b->title),NULL);
 			clr_white(txt);
 			txt1=convert_to_entity_string(txt);
-			add_to_str(&data, &l, cast_uchar "    <DT><H3>");
+			add_to_str(&data, &l, "    <DT><H3>");
 			add_to_str(&data, &l, txt1);
-			add_to_str(&data, &l, cast_uchar "</H3>\n<DL>\n");
+			add_to_str(&data, &l, "</H3>\n<DL>\n");
 			mem_free(txt);
 			mem_free(txt1);
 			depth++;
@@ -840,57 +822,33 @@ static void save_bookmarks(struct session *ses)
 		else
 		{
 			unsigned char *txt1, *txt2, *txt11;
-			txt1=convert_string(ct,b->title,strlen(cast_const_char b->title),NULL);
+			txt1=convert_string(ct,b->title,strlen(b->title),NULL);
 			clr_white(txt1);
-			txt2=convert_string(ct,b->url,strlen(cast_const_char b->url),NULL);
+			txt2=convert_string(ct,b->url,strlen(b->url),NULL);
 			clr_white(txt2);
 			txt11=convert_to_entity_string(txt1);
-			add_to_str(&data, &l, cast_uchar "    <DT><A HREF=\"");
+			add_to_str(&data, &l, "    <DT><A HREF=\"");
 			add_to_str(&data, &l, txt2);
-			add_to_str(&data, &l, cast_uchar "\">");
+			add_to_str(&data, &l, "\">");
 			add_to_str(&data, &l, txt11);
-			add_to_str(&data, &l, cast_uchar "</A>\n");
+			add_to_str(&data, &l, "</A>\n");
 			mem_free(txt1);
 			mem_free(txt2);
 			mem_free(txt11);
 		}
 	}
-	for (a=0;a<depth;a++)add_to_str(&data, &l,cast_uchar "</DL>\n");
-	add_to_str(&data, &l, cast_uchar 
+	for (a=0;a<depth;a++)add_to_str(&data, &l,"</DL>\n");
+	add_to_str(&data, &l,
 	"</DL><P>\n"
 	"</HTML>\n"
 	);
-	err = write_to_config_file(bookmarks_file, data);
+	write_to_config_file(bookmarks_file, data);
 	mem_free(data);
-	if (!err) {
-		bookmark_ld.modified=0;
-	} else {
-		if (ses) {
-			unsigned char *f = stracpy(bookmarks_file);
-			msg_box(ses->term, getml(f, NULL), TEXT_(T_BOOKMARK_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_BOOKMARK_FILE), cast_uchar " ", f, cast_uchar ": ", get_err_msg(err), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
-		}
-	}
-
-	EINTRLOOP(rs, stat(cast_const_char bookmarks_file, &bookmarks_st));
-	if (rs)
-		memset(&bookmarks_st, 0, sizeof bookmarks_st);
+	bookmark_ld.modified=0;
 }
 
 void menu_bookmark_manager(struct terminal *term,void *fcp,struct session *ses)
 {
-	struct stat st;
-	int rs;
-	EINTRLOOP(rs, stat(cast_const_char bookmarks_file, &st));
-	if (!rs &&
-	    (st.st_ctime != bookmarks_st.st_ctime ||
-	     st.st_mtime != bookmarks_st.st_mtime ||
-	     st.st_size != bookmarks_st.st_size)) {
-		if (!test_list_window_in_use(&bookmark_ld, NULL)) {
-			free_bookmarks();
-			load_bookmarks(ses);
-			reinit_list_window(&bookmark_ld);
-		}
-	}
 	create_list_window(&bookmark_ld,&bookmarks,term,ses);
 }
 
